from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Defines an *experimental* gauge-invariant GateSet class and supporting functionality."""

import numpy as _np

from ..tools import matrixtools as _mt

#import evaltree as _evaltree
from . import gate as _gate
from . import gateset as _gateset

from .verbosityprinter import VerbosityPrinter

SMALL = 1e-10

def _isreal(a):
    return _np.isclose(a.imag, 0.0)

def _angle(x):
    return _np.angle(x) if abs(x) > SMALL else 0.0

class GaugeInvGateSet(object):  #(_collections.OrderedDict):
    """
    Encapsulates a set of gate, state preparation, and POVM effect operations
     in a gauge-invariant manner.

    A GaugeInvGateSet stores a gateset using a *minimal* gauge-invariant
    representation.  This class is *experimental* and not ready for use yet.
    """

    def __init__(self,items=[]):
        """
        Initialize a gauge-invariant gate set, possibly from a list of
          items (used primarily by pickle)
        """
        self.gate_dim = None
        self.E_params = None
        self.D_params = []
        self.B0_params = []
        self.gateLabels = []
        self.storedY0 = None

    def from_gateset(self, gateset, debug=None, verbosity=0, fix=1.0):
        """
        Initialize a gauge-invariant gate set from an existing (gauge-
         variant) GateSet.
        """

        printer = VerbosityPrinter.build_printer(verbosity)

        #This only works for fully-parameterized GateSets so far...
        #assert(gates == True and G0 == True and SPAM == True and SP0 == True)
        #assert(all([isinstance(g, _gate.FullyParameterizedGate)
        #            for g in gateset.values() ]))

        vb = verbosity #shorthand
        self.gate_dim = gateset.get_dimension()
        #self.evstruct = evstruct #eigenvalue structure: a string of "R"s and
        #                         # "C"s for real-pair and conjugate-pair.


        self.gateLabels = list(gateset.keys())
        #gl0 = self.gateLabels[0] #plays a special role

        self.D_params = [None]*len(self.gateLabels)
        Y = [None]*len(self.gateLabels) #List of eigenvector mxs

        #Get parameterization of gate eigenvalues (and get corresponding
        #  eigenvector matrices for later use)
        for i,gl in enumerate(self.gateLabels):
            self.D_params[i], Y[i] = _parameterize_real_gate_mx(gateset[gl],vb,debug)

        #DEBUG
        printer.log("DB: Y0 = %s" % _mt.mx_to_string(Y[0]), 4)
        printer.log("DB: Y0 evals = %s" % _np.linalg.eigvals(Y[0]), 4)
        printer.log("DB: invY0 = %s" %  _mt.mx_to_string(_np.linalg.inv(Y[0])), 4)
        printer.log("DB: rho = %s" % _mt.mx_to_string(gateset.preps[0]), 4)

        #Get parameterization of SPAM pair (assume just a single pair is
        # present for now)
        rho_tilde = _np.dot(_np.linalg.inv(Y[0]), gateset.preps[0])
        E_tilde = _np.dot( _np.transpose(gateset.effects[0]), Y[0] )
        delta0_diag, inv_delta0_diag = \
            _get_delta0_diag(rho_tilde, E_tilde, self.D_params[0], vb)
        delta0 = _np.diag( delta0_diag )
        scaledY0 = _np.dot( Y[0], delta0 )

        #remember the "gauge" of this gateset for going back to a
        # gauge-dependent gateset.
        self.storedY0 = scaledY0

        #DEBUG
        printer.log("DB: Scaled Y0 = %s" %    _mt.mx_to_string(scaledY0))
        printer.log("DB: delta0_diag = %s" %  _mt.mx_to_string(delta0_diag))
        printer.log("DB: rho_tilde = %s" %    _mt.mx_to_string(_np.dot( _np.diag(inv_delta0_diag), rho_tilde)))
        if debug is not None:
            printer.log("DB: BScaled Y0 = " %  _mt.mx_to_string( _np.dot(debug,scaledY0)))

        #Make sure scaling rho_tilde gives vector of all 1s
        #assert( _np.allclose( _np.dot( _np.diag(1.0/delta0_diag), rho_tilde),
        #                      _np.ones(rho_tilde.shape)))

        ET_tilde = _np.dot( _np.transpose(gateset.effects[0]), scaledY0 )
        self.E_params = _get_ET_params(ET_tilde, self.D_params[0])
        printer.log("DB: ET_tilde = %s" % _mt.mx_to_string(ET_tilde))
        if debug is not None:
            printer.log("DB: BScaled ET_tilde = " % _mt.mx_to_string( _np.dot(ET_tilde,debug)))

           #FUTURE: multiple effect-vectors allowed?

        #assume a SPAM pair is present and parameterized (for now) - so
        # parameterize B0j matrices all the same.
        self.B0_params = [None]*(len(self.gateLabels))
        for j,gl in enumerate(self.gateLabels[1:],start=1):
            invYjY0 = _np.dot( _np.linalg.inv(Y[j]), scaledY0 )
            printer.log("DB: invYjY0 = %s" % _mt.mx_to_string(invYjY0))
            self.B0_params[j] = _get_B_params(invYjY0, self.D_params[j],
                                              self.D_params[0], vb)


    def to_gateset(self,verbosity=0):
        """
        Create a gauge-variant GateSet from this gauge-invariant
        representation
        """
        printer = VerbosityPrinter.build_printer(verbosity)

        #We're free to assume some Y0 (~choosing a gauge).  We use the one
        # stored during from_gateset for now.  We *could* choose something
        # else if we wanted (or never stored one) -- it's cols just have to
        # have the correct conjugate-pair structure (given by D_params[0])
        Y0 = self.storedY0
        invY0 = _np.linalg.inv(Y0)

        printer.log(("Y0 = %s") % _mt.mx_to_string(Y0))
        printer.log("Y0 evals = %s" %  _np.linalg.eigvals(Y0))
        printer.log(("inv(Y0) = %s") %  _mt.mx_to_string(invY0))

        #Create the gate set
        gs = _gateset.GateSet()

        #Set E (FUTURE: multiple allowed?)
        ETilde = _get_ETilde_vector( self.E_params, self.D_params[0] )
        Evec = _np.dot(_np.transpose(invY0),ETilde)
        assert( all(_isreal(Evec)) ) # b/c of conjugacy structure
        gs.set_evec( _np.real(Evec) )

        #Set rho
        rhoTilde = _get_rhoTilde(ETilde, self.D_params[0])
        gs.set_rhovec( _np.dot(Y0,rhoTilde) )

        #Set initial gate
        D0 = _constructDj(self.D_params[0])
        mx = _np.dot( Y0, _np.dot(D0, invY0) )
        gs.set_gate(self.gateLabels[0], _gate.FullyParameterizedGate(mx))

        #Set remaining gates
        for i,gl in enumerate(self.gateLabels[1:],start=1):
            mx = _deparameterize_real_gate_mx(self.D_params[i],
                                              self.D_params[0],
                                              self.B0_params[i],
                                              invY0, verbosity)
            gs.set_gate(gl, _gate.FullyParameterizedGate(mx))

        #Set identity vector (store it upon creation?)

        #Return constructed gate set
        return gs


    def from_vector(self, v, gates=True,G0=True,SPAM=True,SP0=True):
        k = 0
        dim = self.gate_dim
        self.E_params = v[k:k+dim]; k += dim
        for i in range(len(self.D_params)):
            self.D_params[i] = v[k:k+dim]; k += dim

        assert(self.B0_params[0] is None) # just a placeholder
        for i in range(1, len(self.B0_params)):
            L = len(self.B0_params[i])
            self.B0_params[i] = v[k:k+L]; k += L
            #TODO: test length L against what it should be

    def to_vector(self, gates=True,G0=True,SPAM=True,SP0=True):
        #concat self.E_params, self.D_params, self.B0_params into one vector
        assert(len(self.E_params) == self.gate_dim)
        assert( all([len(Dp) == self.gate_dim for Dp in self.D_params]))
        to_concat = [self.E_params] + self.D_params + self.B0_params[1:]
        return _np.concatenate(to_concat)

    def num_params(self,gates=True,G0=True,SPAM=True,SP0=True):
        return len(self.to_vector())

    def num_nongauge_params(self,gates=True,G0=True,SPAM=True,SP0=True):
        return self.num_params(gates,G0,SPAM,SP0)

    def num_gauge_params(self,gates=True,G0=True,SPAM=True,SP0=True):
        return 0



def _parameterize_real_gate_mx(real_gate_mx, verbosity, debug):
    """
    Convert the possibly complex eigenvalues of a real matrix into
    the purely real gauge-inv-gateset parameters.
    """

    printer = VerbosityPrinter.build_printer(verbosity)

    #Get (complex) eigenvalues and eigenvectors
    evals,evecs = _np.linalg.eig(real_gate_mx)

    printer.log(" PRE -> Eigenvalues: " % _mt.mx_to_string(evals))
    printer.log(" PRE -> Eigenvecs: " % _mt.mx_to_string(evecs))
    evecs = (1.0+0j)*evecs #to convert evecs to complex (sometimes needed for all-real case)

    #find complex-conjugate (or real degenerate) eigenvalue pairs
    conjugate_pairs = []
    real_pairs = []
    for i,v in enumerate(evals):
        iUsed = [p[0] for p in conjugate_pairs]+[p[1] for p in conjugate_pairs]
        if i in iUsed: continue

        V = evecs[:,i].copy()

        #get all indices j where eval[j] is the conjugate of eval[i] (or degen)
        iClose = [ j for j,w in enumerate(evals[i+1:],start=i+1) \
                       if _np.isclose(v.conj(), w) and j not in iUsed ]
        if len(iClose) == 0: continue #nothing to pair with

        #if multiple "close" indices, choose best one based on evecs
        if _isreal(v):
            if all(_isreal(V)): #then look for another *real* evec
                iReal = [ j for j in iClose if all(_isreal(evecs[:,j])) ]
                if len(iReal) == 0:
                    #raise ValueError("Could not find real pair")
                    continue
                j = iReal[0] #just take first one

                #we treat all degen. eigenvals as conjugate-pairs,
                # so want conjugate-pair, not real eigenvectors:
                W = evecs[:,j].copy()
                evecs[:,i] = (V + 1j*W); evecs[:,j] = (V - 1j*W)
                evecs[:,i] /= _np.linalg.norm(evecs[:,i])
                evecs[:,j] /= _np.linalg.norm(evecs[:,j])

            else: #then look for complex conjugate evec
                iConj = [ j for j in iClose \
                              if all(_np.isclose(evecs[:,j],V.conj())) ]
                if len(iConj) == 0:
                    j = iClose[0]; W = evecs[:,j].copy()
                    printer.log(" AV chk = %s" % _np.linalg.norm( _np.dot(real_gate_mx,V) - v*V))
                    printer.log(" AVc chk= %s" % _np.linalg.norm( _np.dot(real_gate_mx,V.conj()) - v*V.conj()))
                    printer.log(" AW chk= %s"  % _np.linalg.norm( _np.dot(real_gate_mx,W) - v*W))
                    printer.log(" AWc chk= %s" % _np.linalg.norm( _np.dot(real_gate_mx,W.conj()) - v*W.conj()))
                    printer.log("Vnorm = %s"   % _np.linalg.norm(V))
                    printer.log("Wnorm = %s"   % _np.linalg.norm(W))
                    printer.log("nrm = %s"     % _np.linalg.norm(W.real/_np.linalg.norm(W.real)))
                    printer.log(" chk= %s"     %  _np.linalg.norm( W.real/_np.linalg.norm(W.real) - V))

                    printer.log("prod chk = (should be zero) %s" % \
                        _np.linalg.norm( _np.dot(evecs, _np.dot(_np.diag(evals),
                                         _np.linalg.inv(evecs))) - real_gate_mx))

                    evecs[:,i] = W.conj()
                    evecs[:,i] /= _np.linalg.norm(evecs[:,i])

                    printer.log("prod chk2 = (should be zero) %s" % \
                        _np.linalg.norm( _np.dot(evecs, _np.dot(_np.diag(evals),
                                         _np.linalg.inv(evecs))) - real_gate_mx))

                    printer.log("COMPLEX DEGEN PAIR", 5)
                    raise ValueError("Could not find conj pair")

                j = iConj[0] #just take first one

        else: #true complex conjugate pair of eigenvals
            iConj = [ j for j in iClose \
                          if all(_np.isclose(evecs[:,j],V.conj())) ]
            if len(iConj) == 0: raise ValueError("Could not find conj pair")
            j = iConj[0] #just take first one

        #At this point, j == other index to pair
        conjugate_pairs.append( (i,j) )


    #put remaining (non-conj-pair) indices ==> real_pairs
    remaining = [ i for i in range(len(evals)) if \
                      not any([ i in p for p in conjugate_pairs]) ]
    assert(all([ _isreal(evals[i]) for i in remaining ]))

    #sort conjugate pairs by real part then imaginary magnitude
    def cmp_conj_pairs(a, b):
        if evals[a[0]].real > evals[b[0]].real: return 1    # a > b
        elif evals[a[0]].real < evals[b[0]].real: return -1 # a < b
        elif abs(evals[a[0]].imag) > abs(evals[b[0]].imag):
            return 1 # a > b
        elif abs(evals[a[0]].imag) == abs(evals[b[0]].imag):
            return 0
        else:
            return -1
    conjugate_pairs.sort(cmp_conj_pairs, reverse=True)

    #sort remaining (real) eigenvalues
    remaining.sort(key=lambda i: evals[i].real, reverse=True)

    nLeft = len(remaining)
    real_pairs = [(remaining[i],remaining[i+1]) for i in range(0,nLeft-1,2)]

    printer.log("DB: conj pairs = %s" % conjugate_pairs)
    printer.log("DB: real pairs = %s" % real_pairs)

    #create array of eigenvalue parameters, keeping track of the
    # needed eigenvector permutation given the new eigenvalue ordering.
    eval_params = _np.empty( evals.shape, 'd' ) # Note: *not* complex
    permMx = _np.zeros(evecs.shape, 'd')
    k = 0 #running index of current eigenvalue after re-arrangement

    for i,j in conjugate_pairs:
        if evals[i].imag > evals[j].imag: # i => k, j => k+1
            permMx[i,k] = 1.0; permMx[j,k+1] = 1.0
        else: # j => k, i => k+1 (lower index always has greater imag part)
            permMx[i,k] = 1.0; permMx[j,k+1] = 1.0
        eval_params[k] = evals[i].real
        eval_params[k+1] = -abs(evals[i].imag) # neg => conj. pair
        k += 2

    for i,j in real_pairs:
        if evals[i] > evals[j]: # i => k, j => k+1
            permMx[i,k] = 1.0; permMx[j,k+1] = 1.0
        else: # j => k, i => k+1 (lower index always is greater value)
            permMx[i,k] = 1.0; permMx[j,k+1] = 1.0
        eval_params[k] = (evals[i].real + evals[j].real) / 2.0
        eval_params[k+1] = abs(evals[i].real -
                               evals[j].real) / 2.0 # pos => real pair
        k += 2

    if len(remaining) % 2 == 1: #if there's one un-paired (real) eval
        assert(_isreal(evals[remaining[-1]]))
        permMx[remaining[-1],k] = 1.0
        eval_params[k] = evals[remaining[-1]].real

    Y = _np.dot(evecs, permMx)
    printer.log("Parameterizing matrix: %s" % _mt.mx_to_string(real_gate_mx), 4)
    printer.log(" -> Eigenvalues: %s"       % _mt.mx_to_string(evals), 4)
    printer.log(" -> Eigenvectors: %s"      % _mt.mx_to_string(evecs), 4)
    printer.log(" -> Parameters: %s"        % _mt.mx_to_string(eval_params), 4)
    printer.log(" -> Evec Permutation Mx  = %s" % _mt.mx_to_string(permMx), 4)
    printer.log(" -> Y-Mx  = %s"            % _mt.mx_to_string(Y), 4)

    D = _constructDj(eval_params)
    Yi = _np.linalg.inv(Y)
    printer.log(" -> test Y*D*Yi: %s"  %  _mt.mx_to_string(_np.dot(Y,_np.dot(D,Yi))), 4)

    if debug is not None:
        B = debug
        Bi = _np.linalg.inv(debug)
        BY = _np.dot(B,Y)
        BYi = _np.linalg.inv(BY)

        printer.log(" -> debug matrix:%s" % _mt.mx_to_string(_np.dot(B,_np.dot(real_gate_mx,Bi))), 4)
        printer.log(" -> debug Y-Mx  = %s" % _mt.mx_to_string(BY), 4)
        printer.log(" -> debug test BY*D*BYi: %s" % _mt.mx_to_string(_np.dot(BY,_np.dot(D,BYi))), 4)
    printer.log('', 4)

    new_evecs = _np.dot(evecs, permMx)
    return eval_params, new_evecs


def _get_delta0_diag(rho_tilde, E_tilde, D0_params, verbosity):
    """ Trys to set rhoTilde[i] == 1.0 """

    #Note:
    # rho-tilde has form inv(Y0) * rho, where inv(Y0) has conjugate-paired
    # rows as determined by the eigenvalue parameters D0_params and rho
    # is a real column vector.  Thus, inv(Y0) * rho is a complex column
    # vector with the conjugate-pair structure given by D0_params.  Thus,
    # the diagonal of inv(delta0) (or just delta0 since it's a diag mx)
    # will have this same structure, to preserve the conjugate-pair
    # structure of rho-tilde and the B-matrices.
    rho_tilde = rho_tilde.flatten() #so we can index using a single index
    E_tilde = E_tilde.flatten() #so we can index using a single index
    inv_delta0_diag = _np.empty(len(rho_tilde),'complex')

    printer = VerbosityPrinter.build_printer(verbosity)

    printer.log("Finding delta0", 4)
    printer.log(" rho-tilde = %s" % _mt.mx_to_string(rho_tilde), 4)
    printer.log(" E-tilde = %s"   % _mt.mx_to_string(E_tilde), 4)
    printer.log(" D0-params = %s" % _mt.mx_to_string(D0_params), 4)

    #We compute the diagonal of inv(delta0) which makes
    # inv(delta0) * rho_tilde = delta0 * E_tilde (complex case)
    # inv(delta0) * |rho_tilde| = delta0 * |E_tilde| (real case), so
    #  |rho_tilde_element| = delta0**2 * |corresponding_E_tilde_element|
    #  delta0 = sqrt( |rho_tilde_element| / |corresponding_E_tilde_element|)
    # OLD inv(delta0) * inv(Y0) * rho = vector of ones
    for i in range(0,len(D0_params)-1,2):
        _, b = D0_params[i:i+2]
        if b <= 0:
            # complex-conj pair at index i,i+1, so:
            assert(_np.isclose(rho_tilde[i],rho_tilde[i+1].conj()))
            if abs(rho_tilde[i]) < SMALL: # and E_tilde[i] not small??
                printer.warning("(1) scaling near-zero rho_tilde element to 1.0!")
                inv_delta0_diag[i] = 1.0/SMALL
            else:
                inv_delta0_diag[i] = (1.0+0j) / rho_tilde[i] #complex division
            inv_delta0_diag[i+1] = inv_delta0_diag[i].conj()
        else:
            # real pair at index i,i+1, so:
            assert(_isreal(rho_tilde[i]) and _isreal(rho_tilde[i+1]))
            if abs(rho_tilde[i]) < SMALL:
                printer.warning("(2) scaling near-zero rho_tilde element to 1.0!")
                inv_delta0_diag[i] = 1.0/SMALL
            else:
                inv_delta0_diag[i] = 1.0 / rho_tilde[i].real

            if abs(rho_tilde[i+1]) < SMALL:
                printer.warning("(3) scaling near-zero rho_tilde element to 1.0!")
                inv_delta0_diag[i+1] = 1.0/SMALL
            else:
                inv_delta0_diag[i+1] = 1.0 / rho_tilde[i+1].real

    if len(D0_params) % 2 == 1: #then there's an un-paired real eigenvalue
        i = len(D0_params)-1
        assert(_isreal(rho_tilde[i]))
        if abs(rho_tilde[i]) < SMALL:
            printer.log("(4) scaling near-zero rho_tilde element to 1.0!")
            inv_delta0_diag[i] = 1.0/SMALL
        else:
            inv_delta0_diag[i] = 1.0 / rho_tilde[i].real

    delta0_diag = 1.0 / inv_delta0_diag # delta0_diag (could compute directly above?)
    return delta0_diag, inv_delta0_diag


def _get_delta0_diag_mark2(rho_tilde, E_tilde, D0_params, fix, verbosity):
    """ Trys to set rhoTilde[i] == ETilde[i] """

    #Note:
    # rho-tilde has form inv(Y0) * rho, where inv(Y0) has conjugate-paired
    # rows as determined by the eigenvalue parameters D0_params and rho
    # is a real column vector.  Thus, inv(Y0) * rho is a complex column
    # vector with the conjugate-pair structure given by D0_params.  Thus,
    # the diagonal of inv(delta0) (or just delta0 since it's a diag mx)
    # will have this same structure, to preserve the conjugate-pair
    # structure of rho-tilde and the B-matrices.
    rho_tilde = rho_tilde.flatten() #so we can index using a single index
    E_tilde = E_tilde.flatten() #so we can index using a single index
    inv_delta0_diag = _np.empty(len(rho_tilde),'complex')

    printer = VerbosityPrinter.build_printer(verbosity)

    printer.log("Finding delta0", 4)
    printer.log(" rho-tilde = %s" % _mt.mx_to_string(rho_tilde), 4)
    printer.log(" E-tilde = %s" % _mt.mx_to_string(E_tilde), 4)
    printer.log(" D0-params = %s" % _mt.mx_to_string(D0_params), 4)

    #We compute the diagonal of inv(delta0) which makes
    # inv(delta0) * rho_tilde = delta0 * E_tilde (complex case)
    # inv(delta0) * |rho_tilde| = delta0 * |E_tilde| (real case), so
    #  |rho_tilde_element| = delta0**2 * |corresponding_E_tilde_element|
    #  delta0 = sqrt( |rho_tilde_element| / |corresponding_E_tilde_element|)
    # OLD inv(delta0) * inv(Y0) * rho = vector of ones
    for i in range(0,len(D0_params)-1,2):
        _, b = D0_params[i:i+2]
        if b <= 0:
            # complex-conj pair at index i,i+1, so:
            assert(_np.isclose(rho_tilde[i],rho_tilde[i+1].conj()))
            if abs(rho_tilde[i]) < SMALL: # and E_tilde[i] not small??
                printer.warning("(1) scaling near-zero rho_tilde element to 1.0!")
                inv_delta0_diag[i] = 1.0/SMALL
                inv_delta0_diag[i] = 1.0/SMALL
            else:
                inv_delta0_diag[i] = _np.sqrt(E_tilde[i] / rho_tilde[i]) #complex division
            inv_delta0_diag[i+1] = inv_delta0_diag[i].conj()
        else:
            # real pair at index i,i+1, so:
            assert(_isreal(rho_tilde[i]) and _isreal(rho_tilde[i+1]))
            if abs(rho_tilde[i]) < SMALL:
                if abs(E_tilde[i]) < SMALL:
                    #both rho and E are small - ok: leave delta0 == 1.0
                    inv_delta0_diag[i] = fix #1.0 # set sign?
                else:
                    printer.warning("(2) scaling near-zero rho_tilde element to 1.0!")
                    inv_delta0_diag[i] = 1.0/SMALL
            elif abs(E_tilde[i]) < SMALL:
                inv_delta0_diag[i] = fix #1.0 # set sign?
            else:
                inv_delta0_diag[i] = _np.sqrt(abs(E_tilde[i].real / rho_tilde[i].real))
                if (E_tilde[i].real * rho_tilde[i].real < 0 and E_tilde[i] > 0) or \
                   (E_tilde[i].real * rho_tilde[i].real >= 0 and E_tilde[i] < 0):
                    inv_delta0_diag[i] *= -1   # E_tilde[i] should be negative if rho & E
                                               #  are different signs, positive otherwise

            if abs(rho_tilde[i+1]) < SMALL:
                if abs(E_tilde[i+1]) < SMALL:
                    #both rho and E are small - ok: leave delta0 == 1.0
                    inv_delta0_diag[i+1] = fix #1.0 # set sign?
                else:
                    printer.warning("(3) scaling near-zero rho_tilde element to 1.0!")
                    inv_delta0_diag[i+1] = 1.0/SMALL
            elif abs(E_tilde[i+1]) < SMALL:
                inv_delta0_diag[i+1] = fix #1.0 # set sign?
            else:
                inv_delta0_diag[i+1] = _np.sqrt(abs(E_tilde[i+1].real / rho_tilde[i+1].real))
                if (E_tilde[i+1].real * rho_tilde[i+1].real < 0 and E_tilde[i+1] > 0) or \
                   (E_tilde[i+1].real * rho_tilde[i+1].real >= 0 and E_tilde[i+1] < 0):
                    inv_delta0_diag[i+1] *= -1


    if len(D0_params) % 2 == 1: #then there's an un-paired real eigenvalue
        i = len(D0_params)-1
        assert(_isreal(rho_tilde[i]))
        if abs(rho_tilde[i]) < SMALL:
            printer.warning("(4) scaling near-zero rho_tilde element to 1.0!")
            inv_delta0_diag[i] = 1.0/SMALL
        else:
            raise NotImplementedError("TODO - like above")

    #delta0_diag = 1.0 / inv_delta0_diag # delta0_diag (could compute directly above?)
    delta0_diag = _np.array( [ 1.0/x if x != 0.0 else 0.0 for x in inv_delta0_diag ] )
    return delta0_diag, inv_delta0_diag


def _get_ET_params(ET_tilde, D0_params):
    # ET_tilde is a (complex) row-vector with conjugate-pair structure given
    # by D0_params
    ET_tilde = ET_tilde.flatten() #so we can index using a single index
    E_params = _np.empty(len(ET_tilde),'d') #Note: *not* complex

    for i in range(0,len(D0_params)-1,2):
        _, b = D0_params[i:i+2]
        if b <= 0:
            # complex-conj pair at index i,i+1, so:
            assert(_np.isclose(ET_tilde[i],ET_tilde[i+1].conj()))
            E_params[i] = ET_tilde[i].real
            E_params[i+1] = ET_tilde[i].imag # (can be pos or neg)
            #TODO: maybe we divide by abs(b) before setting E_params??
        else:
            # real pair at index i,i+1, so:
            #NAN? assert(_isreal(ET_tilde[i]) and _isreal(ET_tilde[i+1]))
            E_params[i]   = (ET_tilde[i].real + ET_tilde[i+1].real)/2.0
            E_params[i+1] = (ET_tilde[i].real - ET_tilde[i+1].real)/2.0
            #TODO: maybe we divide by abs(b) before setting E_params??

    if len(D0_params) % 2 == 1: #then there's an un-paired real eigenvalue
        i = len(D0_params)-1
        assert(_isreal(ET_tilde[i]))
        E_params[i] = ET_tilde[i].real

    return E_params


def _rowsum(ar, D_params): # for rows where b > 0
    s = 0
    for j in range(0,len(D_params)-1,2): #loop over col-pairs
        _, b = D_params[j:j+2] #cols
        if b <= 0:
            # block is [ [a, a.C], [b, b.C] ], so add a.r + a.i
            s += ar[j].real + ar[j].imag
        else:
            # block is [ [a, b], [c, d] ], so add a + b (normal)
            s += ar[j] + ar[j+1]
    return s


def _get_B_params(invYjY0, Dj_params, D0_params, verbosity):
    #Need to find inv(deltaj) to fix rows ~ such that sum( abs(el) ) == 1
    # B0j == inv(deltaj) * invYjY0  (j != 0)
    # where invYjY0 has special "two-sided" (rows+cols) conjugacy structure
    # given by Dj_params (rows) and D0_params (cols).  Diagonal els of
    # inv(deltaj) will have conjugacy-pair structure of Dj_params so overall
    # structure of B0j is the same as that of invYjY0.  After deltaj element
    # are found, conjugacy-pair structure is used to extract (real) parameters
    # of B0j.
    printer = VerbosityPrinter.build_printer(verbosity)

    assert(len(Dj_params) == len(D0_params))
    inv_deltaj_diag = _np.empty(invYjY0.shape[0],'complex')

    def abssum(ar):
        return sum(map(abs, ar))
    def anglesum(ar):
        return sum(_np.where( _np.absolute(ar) > SMALL, _np.angle(ar), 0) )
    def first_nonzero_angle(ar):
        for x in ar:
            if abs(x) > SMALL:
                return _np.angle(x)
        raise ValueError("No nonzero angles found!")

    #DEBUG
    printer.log("DB: invYjY0 (abs,angle) sums = %s" % [ (abssum(invYjY0[i,:]), anglesum(invYjY0[i,:])) for i in range(invYjY0.shape[0])])


    #Record whether the final colum corresponds to a real eigenvalue
    bRealLastCol = (len(D0_params) % 2 == 1) or (D0_params[-1] > 0)

    #get "sums" of invYjY0 rows => choose inv_deltaj_diag accordingly
    for i in range(0,len(Dj_params)-1,2):
        _, b1 = Dj_params[i:i+2] #rows
        if b1 <= 0:
            # sum(row i) should be the conjugate of sum(row i+1)
            abss = abssum(invYjY0[i,:])
            angs = anglesum(invYjY0[i,:])
            assert(_np.isclose( abssum(invYjY0[i+1,:]), abss ))
            #assert(_np.isclose( anglesum(invYjY0[i+1,:]), -angs )) #need mod 2pi...
            nnz = sum([ 1 if abs(x) > SMALL else 0 for x in invYjY0[i,:]]) #num nonzero

            if abss < 1e-10:
                printer.warning("*** Warning *** scaling near-zero B0j row abssum to 1.0!")
                inv_deltaj_diag[i] = 1 / SMALL
            else:
                inv_deltaj_diag[i] = 1.0 / abss   # to make sum of abs(el) == 1
            angle = -angs / nnz                   # to make sum of angles == 0
            mod = 2*_np.pi / nnz

            # can add/subtract "mod" angle and still sum zero, so add/subtract
            # in order to make angle of first nonzero row element close to zero
            fnza = first_nonzero_angle(invYjY0[i,:]) + angle
            #print "DB: first nz angle = ", fnza, " (mod = %g)" % mod
            mod_fctr = 0
            while abs(fnza + mod) < abs(fnza):
                fnza += mod; mod_fctr += 1
            while abs(fnza - mod) < abs(fnza):
                fnza -= mod; mod_fctr -= 1
            angle += mod_fctr * mod
            #print "DB: final fnza = ",fnza, " mod_fctr = ",mod_fctr
            #print "DB: angle = ",angle

            inv_deltaj_diag[i] *= _np.exp(1j*angle)
            inv_deltaj_diag[i+1] = inv_deltaj_diag[i].conj()
        else:
            # sum(row i) and sum(row i+1) should be real
            # scale so that abssum == 1 AND final column is positive (if it's a "real" col)
            abss1 = abssum(invYjY0[i,:])
            abss2 = abssum(invYjY0[i+1,:])

            sign1 = -1 if (bRealLastCol and invYjY0[i,-1] < -SMALL) else 1
            sign2 = -1 if (bRealLastCol and invYjY0[i+1,-1] < -SMALL) else 1

            if abss1 < SMALL:
                printer.warning("*** Warning *** scaling near-zero B0j row abssum to 1.0!")
                inv_deltaj_diag[i] = sign1 * 1/SMALL
            else:
                inv_deltaj_diag[i] = sign1 * 1.0 / abss1

            if abss2 < SMALL:
                printer.warning("*** Warning *** scaling near-zero B0j row abssum to 1.0!")
                inv_deltaj_diag[i+1] = sign2 * 1/SMALL
            else:
                inv_deltaj_diag[i+1] = sign2 * 1.0 / abss2

    if len(Dj_params) % 2 == 1: #then there's an un-paired eigenvalue row
        i = len(Dj_params)-1
        abss = abssum(invYjY0[i,:])
        sign = -1 if (bRealLastCol and invYjY0[i,-1] < 0) else 1

        if abss < 1e-10:
            printer.warning("*** Warning *** scaling near-zero B0j row abssum to 1.0!")
            inv_deltaj_diag[i] = sign * 1/SMALL
        else:
            inv_deltaj_diag[i] = sign * 1.0 / abss

    # Scale invYjY0
    B0j = _np.dot( _np.diag(inv_deltaj_diag), invYjY0 )

    #DEBUG
    printer.log("DB: Bj0 = %s" %  _mt.mx_to_string(B0j), 4)
    printer.log("DB: Bj0 (abs,angle) sums = %s" % [ (abssum(B0j[i,:]), anglesum(B0j[i,:])) for i in range(B0j.shape[0])], 4)

    # Extract (real) parameters from B0j
    L = len(Dj_params)-2 #starting index of final 2x2 block (if one *is* final)
    B0j_params = {} # a dictionary of lists, indexed by two "2x2 block" indices
    for i in range(0,len(Dj_params)-1,2): #loop over row-pairs
        _, b1 = Dj_params[i:i+2] #rows
        for j in range(0,len(D0_params)-1,2): #loop over col-pairs
            _, b2 = D0_params[j:j+2] #cols

            #Each 2x2 square of B0j contains 4 real parameters (after
            # accounting for structure) *except* if j == L, in which case
            # the deltaj-scaling has removed two of these, leaving only 2.
            if b1 <= 0:

                #NOTE: since both b2 cases do exacty the same thing (because
                # a and b lie in the same positions ([i,i] and [i,i+1]) )
                # there's no need to put the code below into the b2 if blocks.
                if j != L: # parameterize a.r, a.i, b.r, b.i
                    B0j_params[i//2,j//2] = \
                        [ B0j[i,j].real, B0j[i,j].imag,
                          B0j[i,j+1].real, B0j[i,j+1].imag ]
                else: # b == 1.0-sum so just parameterize a.r, a.i
                    B0j_params[i//2,j//2] = \
                        [ B0j[i,j].real, B0j[i,j].imag ]
                #TODO: maybe we divide by abs(b) before setting??

                if b2 <= 0:
                    # complex-conj pair rows & cols, so
                    # block is [ [a, b], [b.C, a.C] ]; parameterize a then b.
                    assert(_np.isclose(B0j[i,j], B0j[i+1,j+1].conj()))
                    assert(_np.isclose(B0j[i+1,j], B0j[i,j+1].conj()))
                else:
                    # complex-conj pair rows & real-pair cols, so
                    # block is [ [a, b], [a.C, b.C] ]; parameterize a then b.
                    assert(_np.isclose(B0j[i,j], B0j[i+1,j].conj()))
                    assert(_np.isclose(B0j[i,j+1], B0j[i+1,j+1].conj()))

            else:
                if b2 <= 0:
                    # real-pair rows & complex-conj pair cols,
                    # so block is [ [a, a.C], [b, b.C] ]
                    assert(_np.isclose(B0j[i,j], B0j[i,j+1].conj()))
                    assert(_np.isclose(B0j[i+1,j], B0j[i+1,j+1].conj()))

                    if j != L: # parameterize a.r, a.i, b.r, b.i
                        B0j_params[i//2,j//2] = \
                            [ B0j[i,j].real, B0j[i,j].imag,
                              B0j[i+1,j].real, B0j[i+1,j].imag ]
                    else: # 2|a| == 1-sum1, 2|b| == 1-sum2, so angle(a), angle(b)
                        B0j_params[i//2,j//2] = \
                            [ _angle(B0j[i,j]), _angle(B0j[i+1,j]) ]
                    #TODO: maybe we divide by abs(b) before setting??

                else:
                    # real-pair rows & cols, so block is
                    # [ [a, b], [c, d] ] (all real)
                    assert(_isreal(B0j[i,j]) and _isreal(B0j[i,j+1]) and
                           _isreal(B0j[i+1,j]) and _isreal(B0j[i+1,j+1]))

                    if j != L: # parameterize a, b, c, d
                        B0j_params[i//2,j//2] = \
                            [ B0j[i,j].real, B0j[i,j+1].real,
                              B0j[i+1,j].real, B0j[i+1,j+1].real ]
                    else: # |b| == 1.0-sum1, b>0, |d| == 1.0-sum2, d>0, so just a, c
                        B0j_params[i//2,j//2] = \
                            [ B0j[i,j].real, B0j[i+1,j].real ]
                    #TODO: maybe we divide by abs(b) before setting??

    if len(Dj_params) % 2 == 1: #then there's an un-paired eigenvalue row
        M = len(Dj_params)//2 # index of unpaired block in B0j_params
        i = len(Dj_params)-1
        for j in range(0,len(D0_params)-1,2): #loop over col-pairs (final row)
            _, b2 = D0_params[j:j+2] #cols
            # (note j can never equal L b/c unpaired col exists)

            if b2 <= 0:
                # single-real row & complex-conj pair cols,
                # so block is [a, a.C]; parameterize a
                assert(_np.isclose(B0j[i,j], B0j[i,j+1].conj()))
                B0j_params[M,j//2] = \
                    [ B0j[i,j].real, B0j[i,j].imag ]
                #TODO: maybe we divide by abs(b) before setting??

            else:
                # single-real row & real-pair cols, so block is
                # [a, b] (both real); parameterize a, b
                assert(_isreal(B0j[i,j]) and _isreal(B0j[i,j+1]))
                B0j_params[M,j//2] = \
                    [ B0j[i,j].real, B0j[i,j+1].real ]
                #TODO: maybe we divide by abs(b) before setting??

        j = len(D0_params)-1
        assert(j == i) #since D?_params are the same length
        for i in range(0,len(Dj_params)-1,2): #loop over row-pairs (final col)
            _, b1 = Dj_params[i:i+2] #cols
            # Note j is always the final column here - so the two parameters of
            # each block are set by 1.0-sum of the relevant column.  Thus, there
            # are *no* parameters to add for this final column.

            if b1 <= 0:
                # complex-conj pair rows & single-real col,
                # so block is [[a], [a.C]]; parameterize a
                # (note i can never equal j)
                assert(_np.isclose(B0j[i,j], B0j[i+1,j].conj()))
            else:
                # real-pair rows & single-real col, so block is
                # [[a], [b]] (both real); parameterize a, b
                assert(_isreal(B0j[i,j]) and _isreal(B0j[i+1,j]))

        #Now deal with final diagonal element (single real row & col). The
        # deltaj scaling has set this element to 1-sum as well, so no more
        # parameters are needed.
        #B0j_params[M,M] = [ ] # just so this list exists for loop below?
    else:
        M = None #there is no unpaired block, so no index for it!

    #Now collect B0j_params into a numpy array by concatenating blocks.
    # (This always puts the diagonal blocks into the same locations, so
    #  there's no ambiguity about the size of each segment)
    concat_params = []
    n2x2 = len(Dj_params)//2 # number of 2x2 blocks
    for k in range(n2x2):
        for l in range(n2x2):
            if l == (L+1)//2: assert(len(B0j_params[k,l]) == 2)
            else: assert(len(B0j_params[k,l]) == 4)
            concat_params.extend( B0j_params[k,l] )

    if len(Dj_params) % 2 == 1: #then there's an un-paired eigenvalue
        #Add final row
        for l in range(n2x2):
            assert(len(B0j_params[M,l]) == 2)
            concat_params.extend( B0j_params[M,l] )

        #Add final col (but no parameters!)
        #for k in range(n2x2):
        #    assert(len(B0j_params[k,M]) == 2)
        #    concat_params.extend( B0j_params[k,M] )

    assert( all(_np.isclose(_np.imag(concat_params), 0.0) ))
    return _np.array(concat_params, 'd')


def _constructDj(Dj_params):
    #Construct Dj
    Dj_diag = _np.empty( len(Dj_params), 'complex' )
    for i in range(0,len(Dj_params)-1,2):
        a,b = Dj_params[i:i+2]
        if b <= 0: # complex-conj pair
            Dj_diag[i]   = a-b*1j #a + abs(b)*1j
            Dj_diag[i+1] = a+b*1j #a - abs(b)*1j
        else: # real-pair
            Dj_diag[i]   = a+b
            Dj_diag[i+1] = a-b
    if len(Dj_params) % 2 == 1: # un-paired real eigenvalue
        Dj_diag[-1] = Dj_params[-1]
    return _np.diag(Dj_diag)

def _get_rhoTilde(ETilde, D0_params):
    """ For rhoTilde[i] == 1.0 assumption """
    return _np.ones( (len(D0_params),1), 'complex' )

def _get_rhoTilde_mark2(ETilde, D0_params):
    """ For rhoTilde[i] == ETilde[i] assumption """
    rho_tilde = _np.empty( (len(D0_params),1),'complex')
    for i in range(0,len(D0_params)-1,2):
        _, b = D0_params[i:i+2]
        if b <= 0:
            # complex-conj pair at index i,i+1, so:
            rho_tilde[i] = ETilde[i]
            rho_tilde[i+1] = ETilde[i+1]
        else:
            # real pair at index i,i+1, so:
            rho_tilde[i] = ETilde[i] if ETilde[i] >= 0 else -ETilde[i]
            rho_tilde[i+1] = ETilde[i+1] if ETilde[i+1] >= 0 else -ETilde[i+1]
            #TODO: maybe we divide by abs(b) before setting E_params??

    if len(D0_params) % 2 == 1: #then there's an un-paired real eigenvalue
        rho_tilde[len(D0_params)-1] = Etilde[len(D0_params)-1] if Etilde[len(D0_params)-1] >= 0 else -Etilde[len(D0_params)-1] #pylint: disable=undefined-variable

    return rho_tilde

def _deparameterize_real_gate_mx(Dj_params, D0_params, B0j_params,
                                 invY0, verbosity):
    printer = VerbosityPrinter.build_printer(verbosity)
    assert(len(Dj_params) == len(D0_params))

    def abssum(ar):
        return sum(map(abs,ar))
    def anglesum(ar):
        return sum(_np.where( _np.absolute(ar) > SMALL, _np.angle(ar), 0) )

    #def rowsum(ar, ar_end):
     #   full = list(ar) + list(ar_end)
      #  return _rowsum(full, D0_params)

    #Construct Dj
    Dj = _constructDj(Dj_params)

    #Construct Yj, inv(Yj)
    # B0j := inv(Yj)Y0, so deparameterize B0j then apply inv(Y0)
    B0j = _np.empty( (len(Dj_params),len(Dj_params)), 'complex' )
    L = len(Dj_params)-2 #starting index of final 2x2 block (if one *is* final)

    k = 0 #running index into B0j_params
    for i in range(0,len(Dj_params)-1,2): #loop over row-pairs
        _, b1 = Dj_params[i:i+2] #rows
        for j in range(0,len(D0_params)-1,2): #loop over col-pairs
            _, b2 = D0_params[j:j+2] #cols

            #Each 2x2 square of B0j is specified with 4 real parameters
            # *except* if j == L, in which case our deltaj-scaling has
            # removed two of these, leaving only 2.
            nP = 2 if (j == L) else 4
            params = B0j_params[k:k+nP]
            k += nP

            if b1 <= 0:
                #Get parameters of 2x2 block (which we call a and b below)
                if j == L: #just a parameterized
                    ar,ai = params
                    a = ar + ai*1j
                    abs_b = 1.0 - (abssum(B0j[i,0:L]) + abs(a))
                    angle_b = -1.0 * (anglesum(B0j[i,0:L]) + _angle(a))
                    b = abs_b * _np.exp(1j*angle_b)
                else: # a then b parameterized
                    ar,ai,br,bi = params
                    a = ar + ai*1j
                    b = br + bi*1j

                if b2 <= 0:
                    # complex-conj pair rows & cols, so
                    # block is [ [a, b], [b.C, a.C] ]
                    B0j[i,j:j+2]   = a       , b
                    B0j[i+1,j:j+2] = b.conj(), a.conj()
                else:
                    # complex-conj pair rows & real-pair cols, so
                    # block is [ [a, b], [a.C, b.C] ]
                    B0j[i,j:j+2]   = a       , b
                    B0j[i+1,j:j+2] = a.conj(), b.conj()

            else:
                if b2 <= 0:
                    # real-pair rows & complex-conj pair cols,
                    # so block is [ [a, a.C], [b, b.C] ]

                    #Get parameters of 2x2 block (a and b)
                    if j == L: #a.r+a.i == 1-sum1, b.r+b.i == 1-sum2 ; a.i, b.i
                        angle_a,angle_b = params
                        abs_a = 0.5 * (1.0 - abssum(B0j[i,0:L]))
                        abs_b = 0.5 * (1.0 - abssum(B0j[i+1,0:L]))
                        a = abs_a * _np.exp(1j*angle_a)
                        b = abs_b * _np.exp(1j*angle_b)

                    else: # a then b parameterized
                        ar,ai,br,bi = params
                        a = ar + ai*1j
                        b = br + bi*1j

                    B0j[i,j:j+2]   = a , a.conj()
                    B0j[i+1,j:j+2] = b , b.conj()
                    #TODO: maybe we divide by abs(b) before setting??

                else:
                    # real-pair rows & cols, so block is
                    # [ [a, b], [c, d] ] (all real)

                    #Get parameters of 2x2 block (a, b, c, d)
                    if j == L: # b == 1.0-sum1, d == 1.0-sum2, so just a, c
                        # real part of full row-sum == 1.0
                        a,c = params # AND we've scaled so b, d are positive
                        b = 1.0 - (abssum(B0j[i,0:L]) + abs(a))
                        d = 1.0 - (abssum(B0j[i+1,0:L]) + abs(c))
                    else: # a, b, c, d parameterized
                        a,b,c,d = params

                    B0j[i,j:j+2]   = a , b
                    B0j[i+1,j:j+2] = c , d
                    #TODO: maybe we divide by abs(b) before setting??

    if len(Dj_params) % 2 == 1: #then there's an un-paired eigenvalue row
        i = len(Dj_params)-1
        for j in range(0,len(D0_params)-1,2): #loop over col-pairs
            _, b2 = D0_params[j:j+2] #cols
            params = B0j_params[k:k+2]; k += 2 #always length 2 (j != L always)

            if b2 <= 0:
                # single-real row & complex-conj pair cols,
                # so block is [a, a.C]
                ar,ai = params
                B0j[i,j]   = a + ai*1j
                B0j[i,j+1] = a - ai*1j
                #TODO: maybe we divide by abs(b) before setting??

            else:
                # single-real row & real-pair cols, so block is
                # [a, b] (both real)
                a,b = params
                B0j[i,j]   = a
                B0j[i,j+1] = b
                #TODO: maybe we divide by abs(b) before setting??

        j = len(D0_params)-1
        assert(j == i) #since D?_params are the same length
        for i in range(0,len(Dj_params)-1,2): #loop over row-pairs (final col)
            _, b1 = Dj_params[i:i+2] #cols
            # No params (j == last col, always) - all determined by
            #  rows-wsum-to-one scaling

            if b1 <= 0:
                # complex-conj pair rows & single-real col,
                # so block is [[a], [a.C]]
                abs_a = 1.0 - abssum(B0j[i,0:j])
                angle_a = -anglesum(B0j[i,0:j])
                a = abs_a * _np.exp(1j*angle_a)
                B0j[i,j]   = a
                B0j[i+1,j] = a.conj()
                #TODO: maybe we divide by abs(b) before setting??

            else:
                # real-pair rows & single-real col, so block is
                # [[a], [b]] (both real -- AND positive b/c of scaling)
                a = 1.0 - abssum(B0j[i,0:j])
                b = 1.0 - abssum(B0j[i+1,0:j])
                B0j[i,j]   = a
                B0j[i+1,j] = b
                #TODO: maybe we divide by abs(b) before setting??

        #Now deal with final diagonal element (single real row & col). The
        # deltaj scaling has set the abs-sum of this row to be 1.0 and this
        # element to be positive, so set:
        B0j[j,j] = 1.0 - abssum(B0j[j,0:j])

    invYj = _np.dot( B0j, invY0) # b/c B0j = inv(Yj)Y0
    Yj = _np.linalg.inv(invYj)

    printer.log("De-parameterizing gate:", 4)
    printer.log("B0j = %s" % _mt.mx_to_string(B0j), 4)
    printer.log("Yj = %s" % _mt.mx_to_string(Yj), 4)
    printer.log("invYj = %s" % _mt.mx_to_string(invYj), 4)
    printer.log("Dj = %s" % _mt.mx_to_string(Dj))
    printer.log("mx = %s" % _mt.mx_to_string( _np.dot(Yj, _np.dot(Dj, invYj))), 4)

    #Construct gate
    mx = _np.dot(Yj, _np.dot(Dj, invYj))
    assert(all(_isreal(mx.flatten())))

    return _np.real(mx)


def _get_ETilde_vector( E_params, D0_params ):
    E_tilde = _np.empty( (len(E_params),1),'complex')

    for i in range(0,len(D0_params)-1,2):
        _, b = D0_params[i:i+2]
        if b <= 0:
            # complex-conj pair at index i,i+1, so:
            Er,Ei = E_params[i:i+2]
            E_tilde[i]   = Er + Ei*1j
            E_tilde[i+1] = Er - Ei*1j
            #TODO: maybe we divide by abs(b) before setting E_params??
        else:
            # real pair at index i,i+1, so:
            E1,E2 = E_params[i:i+2]
            E_tilde[i]   = E1 + E2
            E_tilde[i+1] = E1 - E2
            #TODO: maybe we divide by abs(b) before setting E_params??

    if len(D0_params) % 2 == 1: #then there's an un-paired real eigenvalue
        E_tilde[len(D0_params)-1] = E_params[len(D0_params)-1]

    return E_tilde



from __future__ import division, print_function, absolute_import, unicode_literals
import pickle as _pickle
import matplotlib
import matplotlib.pyplot as _plt
import tempfile          as _tempfile
import shutil            as _shutil

from contextlib import contextmanager

import os

@contextmanager
def temp_backend(backend):
    oldBackend = matplotlib.get_backend()
    _plt.switch_backend(backend)
    yield # User code goes here
    _plt.switch_backend(oldBackend)

class ReportFigure(object):
    """
    Encapsulates a single report figure.

    Essentially a matplotlib figure, but with better persistence.  In
    particular, the figure can be held in memory or on disk and saved
    in different formats independently of the state of the "matplotlib
    cloud".
    """

    def __init__(self, axes, extraInfo=None, imageformats=['png', 'pdf']):
        """
        Create a new ReportFigure.

        Parameters
        ----------
        axes : Axes
           Matplotlib axes of the figure.

        extraInfo : object
           Any extra information you want to
           associate with this figure.
        """

        self.extraInfo   = extraInfo
        self.pickledAxes = _pickle.dumps(axes)

        self.tempFileDict = { }

        for fmt in imageformats:
            tf = _tempfile.TemporaryFile()
            _plt.savefig(tf, format=fmt, bbox_extra_artists=(axes,),
                         bbox_inches='tight') #need extra artists otherwise
            tf.seek(0) #rewind to beginning of file
            self.tempFileDict[fmt] = bytearray(tf.read())
            tf.close()


    def _save_axes(self, filename):
        axes = _pickle.loads(self.pickledAxes)
          #this creates a new (current) figure in matplotlib
        curFig = _plt.gcf() # gcf == "get current figure"
        curFig.callbacks.callbacks = {}
          # initialize fig's CallbackRegistry, which doesn't
          # unpickle properly in matplotlib 1.5.1 (bug?)
        _plt.savefig(filename, bbox_extra_artists=(axes,),
                     bbox_inches='tight') #need extra artists otherwise
                                          #axis labels get clipped
        _plt.close(curFig) # closes the figure created by unpickling

    def save_to(self, filename):
        if filename is not None and len(filename) > 0:
            try:
                self._save_axes(filename)
            except Exception as e:
                print('Warning: unpickling to save figure %s failed:\n  %s' % (filename, str(e)))
                ext = os.path.splitext(filename)[1][1:] # remove .
                if ext in self.tempFileDict:
                    tf = _tempfile.NamedTemporaryFile()
                    tf.write(self.tempFileDict[ext])
                    _shutil.copy2(tf.name, filename)
                    tf.close()
                    print("  --> Successfully used cached %s format" % ext)
                else:
                    print('Extension not in cached files and unpickling failed. Trying experimental backend switching. Your machine may catch fire..')
                    # Subprocess didn't work without auxillary file (cannot import parent module '') -> we weren't desperate enough to do the auxillary file, so no promises that works either
                    # multiprocessing didn't work (it used the same matplotlib import/backend :( )
                    # tempfile with jpeg didn't work (we needed diverse formats)
                    try:
                        with temp_backend('agg'):
                            self._save_axes(filename)
                    except:
                        raise ValueError("ReportFigure unpickling error!  This " +
                                         "could be caused by using matplotlib or " +
                                         "pylab magic functions ('%pylab inline' or " +
                                         "'%matplotlib inline') within an iPython " +
                                         "notebook, so if you used either of these " +
                                         "please remove it and all should be well.")

    def set_extra_info(self, extraInfo):
        self.extraInfo = extraInfo

    def get_extra_info(self):
        return self.extraInfo

    def check(self):
        pass
        #axes = _pickle.loads(self.pickledAxes) #pylint: disable=unused-variable
          #this creates a new (current) figure in matplotlib
        #curFig = _plt.gcf() # gcf == "get current figure"
        #curFig.callbacks.callbacks = {} # initialize fig's CallbackRegistry...
        #_plt.close(curFig) # closes the figure created by unpickling



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Functions for reducing the number of required fiducial pairs for analysis."""

import numpy      as _np
import itertools  as _itertools
import random     as _random
import scipy.misc as _spmisc
from ..construction import gatestringconstruction as _gsc
from ..tools        import remove_duplicates      as _remove_duplicates

from ..             import objects as _objs

def _nCr(n,r):
    """Number of combinations of r items out of a set of n.  Equals n!/(r!(n-r)!)"""
    #f = _math.factorial; return f(n) / f(r) / f(n-r)
    return _spmisc.comb(n,r)

def _random_combination(indices_tuple, r):
    """
    Random selection from itertools.combinations(indices_tuple, r)
      from http://docs.python.org/2/library/itertools.html#recipes 
    """
    n = len(indices_tuple)
    iis = sorted(_random.sample(range(n), r))
    return tuple(indices_tuple[i] for i in iis)

def find_sufficient_fiducial_pairs(targetGateset, prepStrs, effectStrs, germList,
                                   testLs=(256,2048), spamLabels="all", tol=0.75,
                                   searchMode="sequential", nRandom=100, seed=None,
                                   verbosity=0, testPairList=None, memLimit=None,
                                   minimumPairs=1):
    """
    Finds a (global) set of fiducial pairs that are amplificationally complete.

    A "standard" set of GST gate sequences consists of all sequences of the form:

    statePrep + prepFiducial + germPower + measureFiducial + measurement
    
    This set is typically over-complete, and it is possible to restrict the 
    (prepFiducial, measureFiducial) pairs to a subset of all the possible 
    pairs given the separate `prepStrs` and `effectStrs` lists.  This function
    attempts to find a set of fiducial pairs that still amplify all of the
    gate set's parameters (i.e. is "amplificationally complete").  The test
    for amplification is performed using the two germ-power lengths given by
    `testLs`, and tests whether the magnitudes of the Jacobian's singular
    values scale linearly with the germ-power length.

    In the special case when `testPairList` is not None, the function *tests*
    the given set of fiducial pairs for amplificational completeness, and 
    does not perform any search.

    Parameters
    ----------
    targetGateset : GateSet
        The target gateset used to determine amplificational completeness.

    prepStrs, effectStrs, germList : list of GateStrings
        The (full) fiducial and germ gate sequences.

    testLs : (L1,L2) tuple of ints, optional
        A tuple of integers specifying the germ-power lengths to use when
        checking for amplificational completeness.

    spamLabels : list or "all", optional
        A list of the SPAM labels to consider when checking for completeness.
        Usually this should be left as the special (and default) value "all",
        which considers all of the SPAM labels.

    tol : float, optional
        The tolerance for the fraction of the expected amplification that must
        be observed to call a parameter "amplified".

    searchMode : {"sequential","random"}, optional
        If "sequential", then all potential fiducial pair sets of a given length
        are considered in sequence before moving to sets of a larger size.  This
        can take a long time when there are many possible fiducial pairs.
        If "random", then only `nRandom` randomly chosen fiducial pair sets are
        considered for each set size before the set is enlarged.

    nRandom : int, optional
        The number of random-pair-sets to consider for a given set size.

    seed : int, optional
        The seed to use for generating random-pair-sets.
        
    verbosity : int, optional
        How much detail to print to stdout.

    testPairList : list or None, optional
        If not None, a list of (iRhoStr,iEffectStr) tuples of integers,
        specifying a list of fiducial pairs (indices are into `prepStrs` and
        `effectStrs`, respectively).  These pairs are then tested for 
        amplificational completeness and the number of amplified parameters
        is printed to stdout.  (This is a special debugging functionality.)

    memLimit : int, optional
        A memory limit in bytes.

    minimumPairs : int, optional
        The minimium number of fiducial pairs to try (default == 1).  Set this
        to integers larger than 1 to avoid trying pair sets that are known to 
        be too small.

    Returns
    -------
    list
        A list of (iRhoStr,iEffectStr) tuples of integers, specifying a list
        of fiducial pairs (indices are into `prepStrs` and `effectStrs`).
    """
    printer = _objs.VerbosityPrinter.build_printer(verbosity)
    #trim LSGST list of all f1+germ^exp+f2 strings to just those needed to get full rank jacobian. (compressed sensing like)

    #tol = 0.5 #fraction of expected amplification that must be observed to call a parameter "amplified"
    if spamLabels == "all":
        spamLabels = targetGateset.get_spam_labels()

    nGatesetParams = targetGateset.num_params()

    #Compute all derivative info: get derivative of each <E_i|germ^exp|rho_j> where i = composite EVec & fiducial index and j similar
    def get_derivs(L):
        dP = _np.empty( (len(germList),len(spamLabels),len(effectStrs)*len(prepStrs), nGatesetParams) )
           #indexed by [iGerm,iSpamLabel,iFiducialPair,iGatesetParam] : gives d(<SP|f0+exp_iGerm+f1|AM>)/d(iGatesetParam)

        for iGerm,germ in enumerate(germList):
            expGerm = _gsc.repeat_with_max_length(germ,L) # could pass exponent and set to germ**exp here
            lst = _gsc.create_gatestring_list("f0+expGerm+f1", f0=prepStrs, f1=effectStrs,
                                                         expGerm=expGerm, order=('f0','f1'))
            evTree = targetGateset.bulk_evaltree(lst)
            blkSz = None

            if memLimit is not None:
                #nSpam = targetGateset.num_preps() * targetGateset.num_effects()
                dim = targetGateset.get_dimension()
                nParams = targetGateset.num_params()
                memEstimate = 8.0*len(evTree)*dim**2*nParams #*nSPAM
                if memEstimate > memLimit:
                    printer.log("Germ %d/%d: Memory estimate = %.1fGB > %.1fGB" % \
                        (iGerm+1,len(germList),memEstimate/(1024.0**3),memLimit/(1024.0**3)))
                    if memEstimate/nParams < memLimit:
                        blkSz = int(nParams * memLimit / memEstimate)
                        printer.log(" --> Setting max block size = %s" % blkSz)
                    else:
                        # Tree and deriv splitting method
                        blkSz = 1
                        maxTreeLength = len(evTree) * memLimit / (memEstimate / nParams)
                        evTree.split(maxTreeLength) # Often fails because maxTreeLength is too small
                        printer.log(" --> Setting max subtree size = %s" % maxTreeLength)
                        evTree.print_analysis()

            dprobs = targetGateset.bulk_dprobs(evTree, wrtBlockSize=blkSz)
            for iSpamLabel,spamLabel in enumerate(spamLabels):
                dP[iGerm, iSpamLabel, :,:] = dprobs[spamLabel]
        return dP

    def get_number_amplified(M0,M1,L0,L1,verb):
        printer = _objs.VerbosityPrinter.build_printer(verb)
        L_ratio = float(L1)/float(L0)
        try:
            s0 = _np.linalg.svd(M0, compute_uv=False)
            s1 = _np.linalg.svd(M1, compute_uv=False)
        except:
            printer.warning("SVD error!!"); return 0
            #SVD did not converge -> just say no amplified params...

        numAmplified = 0
        printer.log("Amplified parameter test: matrices are %s and %s." % (M0.shape, M1.shape), 4)
        printer.log("Index : SV(L=%d)  SV(L=%d)  AmpTest ( > %g ?)" % (L0,L1,tol), 4)
        for i,(v0,v1) in enumerate(zip(sorted(s0,reverse=True),sorted(s1,reverse=True))):
            if abs(v0) > 0.1 and (v1/v0)/L_ratio > tol:
                numAmplified += 1
                printer.log("%d: %g  %g  %g YES" % (i,v0,v1, (v1/v0)/L_ratio ), 4)
            printer.log("%d: %g  %g  %g NO" % (i,v0,v1, (v1/v0)/L_ratio ), 4)
        return numAmplified

    #rank = len( [v for v in s if v > 0.001] )


    printer.log("------  Fiducial Pair Reduction --------")

    L0 = testLs[0]; dP0 = get_derivs(L0)
    L1 = testLs[1]; dP1 = get_derivs(L1)
    fullTestMx0 = dP0.view(); fullTestMx0.shape = ( (len(germList)*len(spamLabels)*len(prepStrs)*len(effectStrs), nGatesetParams) )
    fullTestMx1 = dP1.view(); fullTestMx1.shape = ( (len(germList)*len(spamLabels)*len(prepStrs)*len(effectStrs), nGatesetParams) )

    #Get number of amplified parameters in the "full" test matrix: the one we get when we use all possible fiducial pairs
    if testPairList is None:
        maxAmplified = get_number_amplified(fullTestMx0, fullTestMx1, L0, L1, verbosity+1)
        printer.log("maximum number of amplified parameters = %s" % maxAmplified)

    #Loop through fiducial pairs and add all derivative rows (1 x nGatesetParams) to test matrix
    # then check if testMatrix has full rank ( == nGatesetParams)

    nPossiblePairs = len(prepStrs)*len(effectStrs)
    allPairIndices = list(range(nPossiblePairs))

    nRhoStrs, nEStrs = len(prepStrs), len(effectStrs)
    germFctr = len(spamLabels)*len(prepStrs)*len(effectStrs); nGerms = len(germList)
    spamLabelFctr = len(prepStrs)*len(effectStrs); nSpamLabels = len(spamLabels)
    gateStringIndicesForPair = []
    for i in allPairIndices:
        indices = [ iGerm*germFctr + iSpamLabel*spamLabelFctr + i  for iGerm in range(nGerms) for iSpamLabel in range(nSpamLabels) ]
        gateStringIndicesForPair.append(indices)

    if testPairList is not None: #special mode for testing/debugging single pairlist
        gateStringIndicesForPairs = []
        for iRhoStr,iEStr in testPairList:
            gateStringIndicesForPairs.extend( gateStringIndicesForPair[iRhoStr*nEStrs + iEStr] )
        testMx0 = _np.take( fullTestMx0, gateStringIndicesForPairs, axis=0 )
        testMx1 = _np.take( fullTestMx1, gateStringIndicesForPairs, axis=0 )
        nAmplified = get_number_amplified(testMx0, testMx1, L0, L1, verbosity)
        printer.log("Number of amplified parameters = %s" % nAmplified)
        return None

    bestAmplified = 0
    for nNeededPairs in range(minimumPairs,nPossiblePairs):
        printer.log("Beginning search for a good set of %d pairs (%d pair lists to test)" % \
                (nNeededPairs,_nCr(nPossiblePairs,nNeededPairs)))

        bestAmplified = 0
        if searchMode == "sequential":
            pairIndicesToIterateOver = _itertools.combinations(allPairIndices, nNeededPairs)

        elif searchMode == "random":
            _random.seed(seed)  # ok if seed is None
            nTotalPairCombos = _nCr(len(allPairIndices), nNeededPairs)
            if nRandom < nTotalPairCombos:
                pairIndicesToIterateOver = [ _random_combination(allPairIndices, nNeededPairs) for i in range(nRandom) ]
            else:
                pairIndicesToIterateOver = _itertools.combinations(allPairIndices, nNeededPairs)

        for pairIndicesToTest in pairIndicesToIterateOver:
            gateStringIndicesForPairs = []
            for i in pairIndicesToTest:
                gateStringIndicesForPairs.extend( gateStringIndicesForPair[i] )
            testMx0 = _np.take( fullTestMx0, gateStringIndicesForPairs, axis=0 )
            testMx1 = _np.take( fullTestMx1, gateStringIndicesForPairs, axis=0 )
            nAmplified = get_number_amplified(testMx0, testMx1, L0, L1, verbosity)
            bestAmplified = max(bestAmplified, nAmplified)
            if printer.verbosity > 1:
                ret = []
                for i in pairIndicesToTest:
                    iRhoStr = i // nEStrs
                    iEStr   = i - iRhoStr*nEStrs
                    ret.append( (iRhoStr,iEStr) )
                printer.log("Pair list %s ==> %d amplified parameters" % (" ".join(map(str,ret)), nAmplified))

            if nAmplified == maxAmplified:
                ret = []
                for i in pairIndicesToTest:
                    iRhoStr = i // nEStrs
                    iEStr   = i - iRhoStr*nEStrs
                    ret.append( (iRhoStr,iEStr) )
                return ret

    printer.log(" --> Highest number of amplified parameters was %d" % bestAmplified)

    #if we tried all the way to nPossiblePairs-1 and no success, just return all the pairs, which by definition will hit the "max-amplified" target
    listOfAllPairs = [ (iRhoStr,iEStr)
                       for iRhoStr in range(nRhoStrs)
                       for iEStr in range(nEStrs) ]
    return listOfAllPairs



def find_sufficient_fiducial_pairs_per_germ(targetGateset, prepStrs, effectStrs,
                                            germList, spamLabels="all",
                                            searchMode="sequential", constrainToTP=True,
                                            nRandom=100, seed=None, verbosity=0,
                                            memLimit=None):
    """
    Finds a per-germ set of fiducial pairs that are amplificationally complete.

    A "standard" set of GST gate sequences consists of all sequences of the form:

    statePrep + prepFiducial + germPower + measureFiducial + measurement
    
    This set is typically over-complete, and it is possible to restrict the 
    (prepFiducial, measureFiducial) pairs to a subset of all the possible 
    pairs given the separate `prepStrs` and `effectStrs` lists.  This function
    attempts to find sets of fiducial pairs, one set per germ, that still
    amplify all of the gate set's parameters (i.e. is "amplificationally
    complete").  For each germ, a fiducial pair set is found that amplifies
    all of the "parameters" (really linear combinations of them) that the
    particular germ amplifies.  

    To test whether a set of fiducial pairs satisfies this condition, the
    sum of projectors `P_i = dot(J_i,J_i^T)`, where `J_i` is a matrix of the
    derivatives of each of the selected (prepFiducial+germ+effectFiducial)
    sequence probabilities with respect to the i-th germ eigenvalue (or
    more generally, amplified parameter), is computed.  If the fiducial-pair
    set is sufficient, the rank of the resulting sum (an operator) will be
    equal to the total (maximal) number of parameters the germ can amplify.

    Parameters
    ----------
    targetGateset : GateSet
        The target gateset used to determine amplificational completeness.

    prepStrs, effectStrs, germList : list of GateStrings
        The (full) fiducial and germ gate sequences.

    spamLabels : list or "all", optional
        A list of the SPAM labels to consider when checking for completeness.
        Usually this should be left as the special (and default) value "all",
        which considers all of the SPAM labels.

    searchMode : {"sequential","random"}, optional
        If "sequential", then all potential fiducial pair sets of a given length
        are considered in sequence (per germ) before moving to sets of a larger
        size.  This can take a long time when there are many possible fiducial
        pairs.  If "random", then only `nRandom` randomly chosen fiducial pair
        sets are considered for each set size before the set is enlarged.

    constrainToTP : bool, optional
        Whether or not to consider non-TP parameters the the germs amplify.  If
        the fiducal pairs will be used in a GST estimation where the gate set is
        constrained to being trace-preserving (TP), this should be set to True.

    nRandom : int, optional
        The number of random-pair-sets to consider for a given set size.

    seed : int, optional
        The seed to use for generating random-pair-sets.
        
    verbosity : int, optional
        How much detail to print to stdout.

    memLimit : int, optional
        A memory limit in bytes.

    Returns
    -------
    dict
        A dictionary whose keys are the germ gate strings and whose values are
        lists of (iRhoStr,iEffectStr) tuples of integers, each specifying the
        list of fiducial pairs for a particular germ (indices are into
        `prepStrs` and `effectStrs`).
    """

    printer = _objs.VerbosityPrinter.build_printer(verbosity)

    if spamLabels == "all":
        spamLabels = targetGateset.get_spam_labels()

    pairListDict = {} # dict of lists of 2-tuples: one pair list per germ

    printer.log("------  Individual Fiducial Pair Reduction --------")
    with printer.progress_logging(1):
        for i,germ in enumerate(germList):

            #Create a new gateset containing static target gates and a
            # special "germ" gate that is parameterized only by it's 
            # eigenvalues (and relevant off-diagonal elements)
            gsGerm = targetGateset.copy()
            gsGerm.set_all_parameterizations("static")
            germMx = gsGerm.product(germ)
            gsGerm.gates["Ggerm"] = _objs.EigenvalueParameterizedGate(
                                                   germMx, True, constrainToTP)

            printer.show_progress(i, len(germList), 
                                  suffix='-- %s germ (%d params)' % 
                                  (germ, gsGerm.num_params()))
            #Debugging
            #print(gsGerm.gates["Ggerm"].evals)
            #print(gsGerm.gates["Ggerm"].params)

            #Get dP-matrix for full set of fiducials, where
            # P_ij = <E_i|germ^exp|rho_j>, i = composite EVec & fiducial index,
            #   j is similar, and derivs are wrt the "eigenvalues" of the germ
            #  (i.e. the parameters of the gsGerm gate set).
            lst = _gsc.create_gatestring_list(
                "f0+germ+f1", f0=prepStrs, f1=effectStrs,
                germ=_objs.GateString(("Ggerm",)), order=('f0','f1'))

            evTree = gsGerm.bulk_evaltree(lst)

             #Memory limit stuff -- TODO: move this within evalTree given
             # some kind of MemoryEstimator object to compute estimates?
            blkSz = None
            if memLimit is not None:
                dim = gsGerm.get_dimension()
                nParams = gsGerm.num_params()
                memEstimate = 8.0*len(evTree)*dim**2*nParams #*nSPAM
                if memEstimate > memLimit:
                    printer.log("Germ %d/%d: Memory estimate = %.1fGB > %.1fGB" % \
                        (i+1,len(germList),memEstimate/(1024.0**3),memLimit/(1024.0**3)))
                    if memEstimate/nParams < memLimit:
                        blkSz = int(nParams * memLimit / memEstimate)
                        printer.log(" --> Setting max block size = %s" % blkSz)
                    else:
                        # Tree and deriv splitting method
                        blkSz = 1
                        maxTreeLength = len(evTree) * memLimit / (memEstimate / nParams)
                        evTree.split(maxTreeLength) # Often fails because maxTreeLength is too small
                        printer.log(" --> Setting max subtree size = %s" % maxTreeLength)
                        evTree.print_analysis()
            
            dprobs = gsGerm.bulk_dprobs(evTree, wrtBlockSize=blkSz)
            dP = _np.concatenate([dprobs[sl] for sl in spamLabels],axis=0)
              #concat along spam labels (just seen as additional fiducials)

            # Construct sum of projectors onto the directions (1D spaces)
            # corresponding to varying each parameter (~eigenvalue) of the
            # germ.  If the set of fiducials is sufficient, then the rank of
            # the resulting operator will equal the number of parameters,
            # indicating that the P matrix is (independently) sensitive to
            # each of the germ parameters (~eigenvalues), which is *all* we
            # want sensitivity to.
            RANK_TOL = 1e-7
            rank = _np.linalg.matrix_rank( _np.dot(dP, dP.T), RANK_TOL )
            if rank < gsGerm.num_params(): # full fiducial set should work!
                raise ValueError("Incomplete fiducial-pair set!")

              #Below will take a *subset* of the rows in dprobs[spamLbl]
              # depending on which (of all possible) fiducial pairs
              # are being considered.

            nRhoStrs, nEStrs = len(prepStrs), len(effectStrs)
            nPossiblePairs = len(prepStrs)*len(effectStrs)
            allPairIndices = list(range(nPossiblePairs))

            #Determine which fiducial-pair indices to iterate over
            goodPairList = None; maxRank = 0
            for nNeededPairs in range(gsGerm.num_params(),nPossiblePairs):
                printer.log("Beginning search for a good set of %d pairs (%d pair lists to test)" % \
                                (nNeededPairs,_nCr(nPossiblePairs,nNeededPairs)),2)

                if searchMode == "sequential":
                    pairIndicesToIterateOver = _itertools.combinations(allPairIndices, nNeededPairs)

                elif searchMode == "random":
                    _random.seed(seed)  # ok if seed is None
                    nTotalPairCombos = _nCr(len(allPairIndices), nNeededPairs)
                    if nRandom < nTotalPairCombos:
                        pairIndicesToIterateOver = [ _random_combination(allPairIndices, nNeededPairs) for i in range(nRandom) ]
                    else:
                        pairIndicesToIterateOver = _itertools.combinations(allPairIndices, nNeededPairs)

                for pairIndicesToTest in pairIndicesToIterateOver:
    
                    #Get list of pairs as tuples for printing & returning
                    pairList = []
                    for i in pairIndicesToTest:
                        iRhoStr = i // nEStrs; iEStr = i - iRhoStr*nEStrs
                        pairList.append( (iRhoStr,iEStr) )
    
                    # Same computation of rank as above, but with only a 
                    # subset of the total fiducial pairs.
                    dps = { sl: _np.take(dprobs[sl],pairIndicesToTest, axis=0)
                            for sl in spamLabels } 
                    dP = _np.concatenate([dps[sl] for sl in spamLabels],axis=0)
                    rank = _np.linalg.matrix_rank( _np.dot(dP, dP.T), RANK_TOL )
                    maxRank = max(maxRank,rank)
    
                    printer.log("Pair list %s ==> %d of %d amplified parameters"
                                % (" ".join(map(str,pairList)), rank,
                                   gsGerm.num_params()), 3)
    
                    if rank == gsGerm.num_params():
                        printer.log("Found a good set of %d pairs: %s" % \
                                        (nNeededPairs," ".join(map(str,pairList))),2)
                        goodPairList = pairList
                        break

                if goodPairList is not None:
                    break #exit another loop level if a solution was found

            if goodPairList is not None:
                pairListDict[germ] = goodPairList # add to final list of per-germ pairs
            else:
                #we tried all the way to nPossiblePairs-1 and no success,
                # just return all the pairs
                printer.log(" --> Highest number amplified = %d of %d" %
                            (maxRank, gsGerms.num_params()))
                listOfAllPairs = [ (iRhoStr,iEStr)
                                   for iRhoStr in range(nRhoStrs)
                                   for iEStr in range(nEStrs) ]
                pairListDict[germ] = listOfAllPairs

    return pairListDict





#def _old_TestPair(targetGateset, fiducialList, germList, L, testPairList, spamLabels="all"):
#
#    if spamLabels == "all":
#        spamLabels = targetGateset.get_spam_labels()
#
#    dprobs = []
#    for iGerm,germ in enumerate(germList):
#        expGerm = _gsc.repeat_with_max_length(germ,L)
#        lst = _gsc.create_gatestring_list("f0+expGerm+f1", f0=fiducialList, f1=fiducialList,
#                                        expGerm=expGerm, order=('f0','f1'))
#        evTree = targetGateset.bulk_evaltree(lst)
#        dprobs.append( targetGateset.bulk_dprobs(evTree) )
#
#    nGatesetParams = targetGateset.num_params()
#    testMatrix = _np.empty( (0,nGatesetParams) )
#    for (i0,i1) in testPairList: #[(0,0),(1,0),(2,3),(4,5)]:
#        iCmp = i0*len(fiducialList) + i1 #composite index of (f0,f1) in dprobs[iGerm][spamLabel]
#
#        for iGerm,germ in enumerate(germList):
#            for spamLabel in spamLabels:
#                testMatrix = _np.concatenate( (testMatrix, dprobs[iGerm][spamLabel][iCmp:iCmp+1,:] ), axis=0 )
#
#        U,s,V = _np.linalg.svd(testMatrix)
#        rank = len( [v for v in s if v > 0.001] )
#        sorteds = sorted(s,reverse=True)
#        #print "DEBUG: added (%d,%d): testMx=%s, rank=%d, iCmp=%d, s =\n%s\n" % (i0,i1,testMatrix.shape,rank,iCmp,'\n'.join(map(str,enumerate(sorted(s,reverse=True)))))
#
#    print "Singular values:\n", '\n'.join(map(str,enumerate(sorteds)))
#    #print "normalized 34th singular val = ",sorteds[33]/len(pairList)
#    #print "normalized 35th singular val = ",sorteds[34]/len(pairList)



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" GaugeGroup and derived objects, used primarily in gauge optimization """

import numpy as _np

class GaugeGroup(object):
    def __init__(self):
        pass

    def num_params(self):
        return 0

    def get_element(self, param_vec):
        return self.element()

    def get_initial_params(self):
        return _np.array([],'d')

    class element(object):
        def __init__(self): pass
        def get_transform_matrix(self): return None
        def get_transform_matrix_inverse(self): return None
        def to_vector(self): return _np.array([],'d')
        def from_vector(self,v): pass


class GateGaugeGroup(GaugeGroup):
    def __init__(self, gate):
        self.gate = gate

    def num_params(self):
        return self.gate.num_params()

    def get_element(self, param_vec):
        elgate = self.gate.copy()
        elgate.from_vector(param_vec)
        return self.element(elgate)
    
    def get_initial_params(self):
        return self.gate.to_vector()

    class element(GaugeGroup.element):
        def __init__(self, gate):  
            self.gate = gate
            self._inv_matrix = None
            GaugeGroup.element.__init__(self)

        def get_transform_matrix(self): 
            return _np.asarray(self.gate)

        def get_transform_matrix_inverse(self): 
            if self._inv_matrix is None:
                self._inv_matrix = _np.linalg.inv(_np.asarray(self.gate))
            return self._inv_matrix

        def to_vector(self):
            return self.gate.to_vector()

        def from_vector(self,v):
            self.gate.from_vector(v)
            self._inv_matrix = None



class FullGaugeGroup(GateGaugeGroup):
    def __init__(self, dim):
        from . import gate as _gate #b/c gate.py imports gaugegroup
        gate = _gate.FullyParameterizedGate(_np.identity(dim,'d'))
        GateGaugeGroup.__init__(self, gate)

    class element(GateGaugeGroup.element):
        def __init__(self, gate):
            GateGaugeGroup.element.__init__(self,gate)
            pass #inherits everything it needs

class TPGaugeGroup(GateGaugeGroup):
    def __init__(self, dim):
        from . import gate as _gate #b/c gate.py imports gaugegroup
        gate = _gate.TPParameterizedGate(_np.identity(dim,'d'))
        GateGaugeGroup.__init__(self, gate)

    class element(GateGaugeGroup.element):
        def __init__(self, gate):
            GateGaugeGroup.element.__init__(self,gate)

#        pass #inherits everything it needs

class DiagGaugeGroup(GateGaugeGroup):
    def __init__(self, dim):
        from . import gate as _gate #b/c gate.py imports gaugegroup
        ltrans = _np.identity(dim,'d')
        rtrans = _np.identity(dim,'d')
        baseMx = _np.identity(dim,'d')
        parameterArray = _np.zeros(dim, 'd')
        parameterToBaseIndicesMap = { i: [(i,i)] for i in range(dim) }
        gate = _gate.LinearlyParameterizedGate(baseMx, parameterArray,
                                               parameterToBaseIndicesMap,
                                               ltrans, rtrans, real=True)
        GateGaugeGroup.__init__(self, gate)

    class element(GateGaugeGroup.element):
        def __init__(self, gate):
            GateGaugeGroup.element.__init__(self,gate)

#        pass #inherits everything it needs


class TPDiagGaugeGroup(TPGaugeGroup):
    def __init__(self, dim):
        from . import gate as _gate #b/c gate.py imports gaugegroup
        ltrans = _np.identity(dim,'d')
        rtrans = _np.identity(dim,'d')
        baseMx = _np.identity(dim,'d')
        parameterArray = _np.zeros(dim-1, 'd')
        parameterToBaseIndicesMap = { i: [(i+1,i+1)] for i in range(dim-1) }
        gate = _gate.LinearlyParameterizedGate(baseMx, parameterArray,
                                               parameterToBaseIndicesMap,
                                               ltrans, rtrans, real=True)
        GateGaugeGroup.__init__(self, gate)

    class element(TPGaugeGroup.element):
        def __init__(self, gate):
            TPGaugeGroup.element.__init__(self,gate)

#        pass #inherits everything it needs


class UnitaryGaugeGroup(GateGaugeGroup):
    def __init__(self, dim):
        from . import gate as _gate #b/c gate.py imports gaugegroup
        gate = _gate.LindbladParameterizedGate(None, _np.identity(dim,'d'),
                                               cptp=True, nonHamTerms="none")
        GateGaugeGroup.__init__(self, gate)

    class element(GateGaugeGroup.element):
        def __init__(self, gate):
            GateGaugeGroup.element.__init__(self,gate)



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
""" Defines the OutputData class and supporting functions """

import pickle as _pickle

class OutputData:
    """
    A collection of gatesets, datasets, and parameters for easy saving and loading.
    """
    def __init__(self, init_from_filename=None):
        """ Initialize an OutputData object, either empty by loading data from a file """
        self.datasets = { }
        self.gatesets = { }
        self.parameters = { }
        self.filename = init_from_filename

        if self.filename is not None:
            self.load_from(self.filename)

    def has_gateset(self, gsKey):
        """ Test whether this OutputData contains a gateset named gsKey"""
        return gsKey in self.gatesets

    def has_dataset(self, dsKey):
        """ Test whether this OutputData contains a dataset named dsKey"""
        return dsKey in self.datasets

    def get_gateset(self, gsKey):
        """ Returns the gateset named gsKey.  Returns None if no gateset of that name exists"""
        return self.gatesets.get(gsKey,None)

    def get_dataset(self, dsKey):
        """ Returns the dataset named dsKey.  Returns None if no dataset of that name exists"""
        return self.datasets[dsKey]

    def get_parameter(self, paramName, paramCategory=None):
        """
        Returns the parameter with name paramName, optionally constrained to the category paramCategory.
          If no parameter of that name is found, returns None.
        """
        if paramCategory is None:
            for (_,cat_dict) in self.parameters.items():
                if paramName in cat_dict:
                    return cat_dict[paramName]
        elif paramCategory in self.parameters:
            if paramName in self.parameters[paramCategory]:
                return self.parameters[paramCategory][paramName]
        return None

    def get_parameters(self, paramCategory):
        """ Returns the parameter dictionary corresponding to category paramCategory """
        return self.parameters[paramCategory]

    def set_gateset(self, gsKey, gateset):
        """ Stores a gateset with name gsKey """
        self.gatesets[gsKey] = gateset

    def set_gatesets(self, gsDict):
        """
        Stores gatesets using a dictionary with gateset names given by their dictionary keys
          and the gatesets themselves given by the dictionary values.
        """
        self.gatesets.update(gsDict)

    def set_dataset(self, dsKey, dataset):
        """ Stores a dataset with name dsKey """
        self.datasets[dsKey] = dataset

    def set_datasets(self, dsDict):
        """
        Stores datasets using a dictionary with dataset names given by their dictionary keys
          and the datasets themselves given by the dictionary values.
        """
        self.datasets.update(dsDict)

    def set_parameter(self, paramName, paramValue, paramCategory):
        """ Stores the value paramValue for the parameter named paramName within category paramCategory """
        if paramCategory not in self.parameters:
            self.parameters[paramCategory] = { paramName: paramValue }
        else:
            self.parameters[paramCategory][paramName] = paramValue

    def set_parameters(self, paramCategory, paramDict):
        """ Stores an entire dictionary of parameters paramDict as the category paramCategory """
        self.parameters[paramCategory] = paramDict



    def load_from(self, filename):
        """ Load data from a file.  If filename ends in .gz it will be gzip decompressed. """
        self.filename = filename

        #try:
        if filename.endswith(".gz"):
            import gzip
            file_data = _pickle.load(gzip.open(filename,'rb'))
        else:
            with open(filename, 'rb') as picklefile:
                file_data = _pickle.load(picklefile)
        #except:
        #    raise ValueError("Error loading pickle data from %s" % filename)
        version = file_data.get("gst_output_data_version",0)

        if version == 0:
            for (k,val) in file_data.items():
                if   k == 'gst params': self.set_parameters('algorithm', val)
                elif k == 'data set params': self.set_parameters('dataset', val)
                elif k == 'data set': self.set_dataset('training', val)
                else: self.set_gateset(k, val)

        elif version == 1:
            for (k,val) in file_data.items():
                if   k == 'gatesets': self.gatesets = val
                elif k == 'datasets': self.datasets = val
                elif k == 'parameters': self.parameters = val
                elif k == 'gst_output_data_version': continue
                else: raise ValueError("Invalid version 1 gst output data file: unknown key: %s" % k)

        else:
            raise ValueError("Unknown version (%s) of gst output data file" % version)

    def save_to(self, filename):
        """ Save data to a file.  If filename ends in .gz it will be gzip compressed. """
        dictToSave = { 'gatesets': self.gatesets,
                       'datasets': self.datasets,
                       'parameters': self.parameters,
                       'gst_output_data_version': 1 }
        if filename.endswith(".gz"):
            import gzip as _gzip
            _pickle.dump( dictToSave, _gzip.open(filename, "wb"))
        else:
            with open(filename, 'wb') as picklefile:
                _pickle.dump( dictToSave, picklefile)



    def save(self):
        """ Save data to the same file this OutputData was loaded from """
        if self.filename is None:
            raise ValueError("Cannot save this OutputData object when it wasn't loaded from a file -- must use save_to")
        return self.save_to(self.filename)


#def upgrade_old_data_sets(outputDataWithOldDatasets):
#    newDatasets = { }
#    for key,oldDataset in outputDataWithOldDatasets.datasets.iteritems():
#        newDatasets[key] = _UpgradeOldDataSet(oldDataset)
#    outputDataWithOldDatasets.datasets = newDatasets
#
#def upgrade_old_data_set_pickle(filename):
#    import sys, DataSet, OldDataSet
#
#    currentDataSetModule = sys.modules['DataSet']
#    sys.modules['DataSet'] = OldDataSet  #replace DataSet module with old one so unpickling can work
#    try:      od = OutputData(filename)
#    finally:  sys.modules['DataSet'] = currentDataSetModule
#
#    od.upgrade_old_data_sets(od)
#    od.save_to(filename + ".upd")
#
#    updated_od = OutputData(filename + ".upd")
#    for key,ds in updated_od.datasets.iteritems():
#        print "Upgraded dataset %s, %s" % (key, type(ds))
#
#    print "Successfully updated ==> %s" % (filename + ".upd")



from __future__ import division, print_function, absolute_import, unicode_literals
#*****************************************************************
#    pyGSTi 0.9:  Copyright 2015 Sandia Corporation
#    This Software is released under the GPL license detailed
#    in the file "license.txt" in the top-level pyGSTi directory
#*****************************************************************
"""Functions for generating data with which to evaluate germ sets"""

import numpy as _np

import pygsti
import pygsti.algorithms as alg
import pygsti.construction as constr
import pygsti.objects as obj


def simulate_convergence(germs, prepFiducials, effectFiducials, targetGS,
                         randStr=1e-2, numPertGS=5, maxLengthsPower=8,
                         clickNums=32, numRuns=7, seed=None, randState=None,
                         gaugeOptRatio=1e-3):
    if not isinstance(clickNums, list):
        clickNums = [clickNums]
    if randState is None:
        randState = _np.random.RandomState(seed)

    perturbedGatesets = [targetGS.randomize_with_unitary(scale=randStr,
                                                         randState=randState)
                         for n in range(numPertGS)]

    maxLengths = [0] + [2**n for n in range(maxLengthsPower + 1)]

    expList = constr.make_lsgst_experiment_list(targetGS.gates.keys(),
                                                prepFiducials, effectFiducials,
                                                germs, maxLengths)

    errorDict = {}
    resultDict = {}

    for trueGatesetNum, trueGateset in enumerate(perturbedGatesets):
        for numClicks in clickNums:
            for run in range(numRuns):
                success = False
                failCount = 0
                while not success and failCount < 10:
                    try:
                        ds = constr.generate_fake_data(trueGateset, expList,
                                                       nSamples=numClicks,
                                                       sampleError="binomial",
                                                       randState=randState)

                        result = pygsti.do_long_sequence_gst(
                            ds, targetGS, prepFiducials, effectFiducials,
                            germs, maxLengths, gaugeOptRatio=gaugeOptRatio)

                        errors = [(trueGateset
                                   .frobeniusdist(
                                       alg.gaugeopt_to_target(
                                           estimate, trueGateset,
                                           itemWeights={'spam': 0.0}),
                                       spamWeight=0.0), L)
                                  for estimate, L in 
                                  zip(result.gatesets['iteration estimates'][1:],
                                      result.parameters['max length list'][1:])]

                        resultDict[trueGatesetNum, numClicks, run] = result
                        errorDict[trueGatesetNum, numClicks, run] = errors

                        success = True
                    except Exception as e:
                        failCount += 1
                        if failCount == 10:
                            raise e
                        print(e)

    return obj.GermSetEval(germset=germs, gatesets=perturbedGatesets,
                           resultDict=resultDict, errorDict=errorDict)
